/*
    libporty - collection of random system-dependent network code
    Copyright (C) 2005 gpmi-extension project
    Copyright (C) 2010..2012  Tibor 'Igor2' Palinkas

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA

    Project page: http://repo.hu/projects/libporty
    Author: libporty@igor2.repo.hu
*/

#include <stdio.h>
#include "network.h"
#include "uninit_chain.h"

int P_net_close(P_net_socket s)
{
#ifdef P_NET_CLOSESOCKET
	if (P_net_shutdown(s, P_SHUT_RDWR) != 0)
		return -1;
	return closesocket(s);
#else
	return close(s);
#endif
}


int P_net_set_nonblocking(int sock)
{
#ifdef WIN32
	u_long nonblocking = 1;
#else
	int nonblocking = 1;
#endif
	return P_ioctl_socket(sock, FIONBIO, &nonblocking);
}

int P_net_set_blocking(int sock)
{
#ifdef WIN32
	u_long nonblocking = 0;
#else
	int nonblocking = 0;
#endif

	return P_ioctl_socket(sock, FIONBIO, &nonblocking);
}

/* UDP only! */
int P_net_enable_broadcast(int sock)
{
	/* Enable broadcast */
	unsigned long val = 1;
	return setsockopt(sock, SOL_SOCKET, SO_BROADCAST, (char *)&val , sizeof(val));
}

/* alternative poll() implementations, in case we don't have native poll() */
#ifndef P_NET_POLL

#ifdef P_NET_SELECT
#warning this header should be checked #include <sys/select.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
int P_poll_select(struct P_pollfd *fds, nfds_t nfds, int timeout)
{
	fd_set rd, wr;
	int snfds, n, ret;
	struct timeval to;

	snfds = -1;
	FD_ZERO(&rd);
	FD_ZERO(&wr);
# warning TODO: reproduce odd behavior of poll() returning HUP even if events is 0
	for(n = 0; n < nfds; n++) {
		if ((fds[n].events != 0) && (fds[n].fd >= 0)) {
			if (fds[n].fd >= snfds)
				snfds = fds[n].fd;
			if (fds[n].events & P_POLLIN)
				FD_SET(fds[n].fd, &rd);
			if (fds[n].events & P_POLLOUT)
				FD_SET(fds[n].fd, &wr);
		} else {
			fds[n].revents = 0;
		}
# warning add support for POLLPRI?
	}
	if (snfds == -1) {
		P_usleep(timeout * 1000);
		return 0;
	}

	to.tv_sec  = timeout / 1000;
	to.tv_usec = (timeout - (to.tv_sec * 1000)) * 1000;
	ret = select(snfds + 1, &rd, &wr, NULL, &to);
	if (ret < 0)
		return ret;
	ret = 0;
	for(n = 0; n < nfds; n++) {
		fds[n].revents = 0;
		if (FD_ISSET(fds[n].fd, &rd))
			fds[n].revents |= P_POLLIN;
		if (FD_ISSET(fds[n].fd, &wr))
			fds[n].revents |= P_POLLOUT;
		if (fds[n].revents != 0)
			ret++;
# warning add support for the other standard poll outputs
	}
	return ret;
}

#else
int P_poll_busyloop(struct P_pollfd *fds, nfds_t nfds, int timeout)
{
	fprintf(stderr, "P_poll_busyloop() unimplemented yet (TODO)\n");
	abort();
}
#endif

#endif


#warning TODO: do not depend on WIN32, this should be autodetected by scconfig by trying to read() and write() tcp sockets
#ifndef WIN32

/* Sockets can be written and read; async stdio may not be a socket
   thus recv() and send() woudl fail */
P_size_t P_net_read(P_net_socket s, void *buf, P_size_t count)
{
	return read(s, buf, count);
}

P_size_t P_net_write(P_net_socket s, void *buf, P_size_t count)
{
	return write(s, buf, count);
}

#else

/* Sockets can not be handled by read() and write(); we made sure
   async stdio has a real socket and a thread/fork. */
P_size_t P_net_read(P_net_socket s, void *buf, P_size_t count)
{
	return recv(s, buf, count, 0);
}

P_size_t P_net_write(P_net_socket s, void *buf, P_size_t count)
{
	return send(s, buf, count, 0);
}
#endif


int P_net_printf_minbuff = 1024;
int P_net_printf_maxbuff = 65536;

#ifdef P_HAVE_VDPRINTF
#include <stdarg.h>
P_size_t P_net_printf(P_net_socket s, const char *format, ...)
{
	int ret;
	va_list ap;
	va_start(ap, format);
	ret = vdprintf(s, format, ap);
	va_end(ap);
	return ret;
}
#else
/* we can't use fdopen as an alternative because fclose() would close the
socket. We could hash P_net_socket -> FILE *, but we wouldn't notice
if the fd was replaced meanwhile and the system will most probably
reuse fd numbers. (Or else all libporty calls that could affect what an fd
means would need to maintain the hash). */
#include <stdlib.h>
#include <stdarg.h>

#ifdef P_HAVE_VSNPRINTF
#	ifdef P_HAVE_VSNPRINTF_SAFE
#	define SAFETY 0
#	else
#	define SAFETY 1
#	endif
P_size_t P_net_printf(P_net_socket s, const char *format, ...)
{
	int size;
	char *buff;
	int ret;
	va_list ap;

	for(size = P_net_printf_minbuff; size <= P_net_printf_maxbuff; size = size * 2) {
		buff = malloc(size + SAFETY);
		if (buff == NULL)
			return 0;
		va_start(ap, format);
		ret = vsnprintf(buff, size, format, ap);
		va_end(ap);
		if (ret <= size && ret != -1)
			break;
	}
	if (ret <= size && ret >= 0)
		ret = P_net_write(s, buff, ret);

	free(buff);


	if (ret <= size)
		return ret;
	return -1;
}
#else
/* we don't have anything advanced so allocate a 'large enough' buffer and
check for overflow */
#include <stdlib.h>
#include <stdarg.h>
P_size_t P_net_printf(P_net_socket s, const char *format, ...)
{
	char *buff;
	int len;
	va_list ap;
	va_start(ap, format);

	buff = malloc(P_net_printf_maxbuff);
	len = vsprintf(buff, format, ap);
	if (len >= P_net_printf_maxbuff) {
		fprintf(stderr, "P_net_printf: buffer (%d) too small (need %d); please increase P_net_printf_maxbuff\n", P_net_printf_maxbuff, len);
		abort();
	}
	len = P_net_write(s, buff, len);

	free(buff);
	va_end(ap);
	return len;
}
#endif
#endif


#warning TODO: depends on win32 - better detect these
#ifdef WIN32
int P_net_init(void)
{
	WSADATA w;
	return WSAStartup(MAKEWORD(2,2), &w);
}

int P_net_uninit(void)
{
	P_net_uninit_chain();
	return WSACleanup();
}
#else
#include <signal.h>
int P_net_init(void)
{
	signal(SIGPIPE, SIG_IGN);
	return 0;
}

int P_net_uninit(void)
{
	P_net_uninit_chain();
	return 0;
}
#endif

#ifdef WIN32
P_net_socket P_socket(int af, int protocol, int type)
{
	return WSASocket(af, protocol, type, NULL, 0, 0);
}
#endif
